;; Generated package description from vcard.el  -*- no-byte-compile: t -*-
(define-package "vcard" "0.2.1.0.20201225.215621" "Package for handling vCard files" '((emacs "27.1")) :url "https://elpa.gnu.org/packages/vcard.html" :authors '(("Eric Abrahamsen" . "eric@ericabrahamsen.net")) :maintainer '("Eric Abrahamsen" . "eric@ericabrahamsen.net"))
